#include <string>
#include <vector>
#include <algorithm>
#include <chrono>
#include <thread>
#include <fstream>
#include <filesystem>
#include <iostream>
#include <cstdio>
#include <dirent.h>
#include <cstring>

#include "server.h"
#include "mode.h"
#include "sleep.h"
#include "commandHandler.h"

#include "audioMixer.h"
#include "a2d_pot.h"
#include "audioManager.h"
#include "buttonHandler.h"
#include "Neotrellis.h"
#include "IODevice.h"

static bool stopping = true;

void serverStop()
{
    stopping = true;
}

bool compareFilesByTimestamp(const std::string &filePath1, const std::string &filePath2)
{
    return std::filesystem::last_write_time(filePath1) < std::filesystem::last_write_time(filePath2);
}

std::vector<std::string> sortFilesByTimestamp(const std::vector<std::string> &fileList)
{
    std::vector<std::string> sortedFileList = fileList;
    std::sort(sortedFileList.begin(), sortedFileList.end(), compareFilesByTimestamp);
    return sortedFileList;
}

std::vector<std::string> getChangesInAscOrder(DIR *dir, const char* dirPath)
{    
    std::vector<std::string> fileList;

    if (dir == nullptr)
    {
        std::cout << "Failed to open directory\n";
        return fileList;
    }

    dirent* entry;
    while ((entry = readdir(dir)) != nullptr)
    {
        if (entry->d_type == DT_REG && entry->d_name[0] != '.')
        {
            fileList.push_back(std::string(dirPath) + "/" + entry->d_name);
        }
    }

    return sortFilesByTimestamp(fileList);
}

std::vector<std::string> readLinesFromFile(const std::string &filePath)
{
    std::ifstream inputFileStream(filePath);
    std::vector<std::string> lines;
    std::string line;
    for (int i = 0; i < 2 && std::getline(inputFileStream, line); ++i)
    {
        lines.push_back(line);
    }
    inputFileStream.close();

    if (lines.size() != 2)
    {
        // unexpected file content, ignore by returning empty
        return std::vector<std::string>();
    }
    return lines;
}

static void initializeDatabase()
{
    const std::string directoryName = "/tmp/changeFeed";

    if (!std::filesystem::exists(directoryName))
    {
        if (std::filesystem::create_directory(directoryName))
        {
            std::cout << "Directory created: " << directoryName << std::endl;
        }
        else
        {
            std::cout << "Failed to create directory: " << directoryName << std::endl;
            exit(1);
        }
    }

    std::cout << "database initialized for server" << std::endl;
}

void handle_terminate_command()
{
    printf("Shutting down\n");

    Pot_stopReading();
    AudioManager_cleanup();
    serverStop();
    AudioMixer_cleanup();
}

void processChange(const std::string &filePath)
{
    // std::cout<< "process change entered"<<std::endl;
    auto lines = readLinesFromFile(filePath);
    if (lines.empty())
    {
        //	std::cout<< "line is enpty"<<std::endl;
        return;
    }

    auto cmd = lines[0];
    auto cmd_value = lines[1];
    if (cmd == "Volume")
    {
        if (cmd_value == "increase")
        {
            volume_increase_command();
        }
        else
        {
            volume_decrease_command();
        }
        std::filesystem::remove(filePath);
    }
    else if (cmd == "Tempo")
    {
        if (cmd_value == "increase")
        {
            tempo_increase_command();
        }
        else
        {
            tempo_decrease_command();
        }
        std::filesystem::remove(filePath);
    }
    else if (cmd == "Button")
    {
        if(cmd_value == "1"){//terminate
            
            printf("terminate: \n");
            std::filesystem::remove(filePath);
            handle_terminate_command();
        }

        printf("setbutton: \n");
        int index = stoi(cmd_value);
        handle_button_command(index);
        std::filesystem::remove(filePath);
    }
    else if (cmd == "Terminate")
    {
        printf("terminate: \n");
        std::filesystem::remove(filePath);
        handle_terminate_command();
    }
    else if(cmd == "VolumeSetter"){
        volume_setter_command(stoi(cmd_value));
        std::filesystem::remove(filePath);
    }
    else if(cmd == "TempoSetter"){
        tempo_setter_command(stoi(cmd_value));
        std::filesystem::remove(filePath);
    }
    else
    {
        std::filesystem::remove(filePath);
    }
}


void processChanges(DIR *dir, const char* dirPath, IODevice *trellis)
{
    auto filePaths = getChangesInAscOrder(dir, dirPath);
    for (const auto &filePath : filePaths)
    {
        processChange(filePath);
    }

    std::vector<int> indexes = trellis->getPressedButtons();
    for(unsigned int i = 0; i < indexes.size(); i++ ){
        handle_button_command(indexes[i]);
    }
}

void serverThreadFunc()
{
    const char* dirPath = "/tmp/changeFeed";
    DIR* dir = opendir(dirPath);
    MockIO trellis;
    while (!stopping)
    {
        processChanges(dir, dirPath, &trellis);
        sleep_x_ms(10);
    }

    closedir(dir);
}

int serverStart()
{

    initializeDatabase();
    //  AudioMixer_init();
    Mode_start();
    //  Pot_startReading();
    //	AudioMixer_startPlayBack();

    stopping = false;
    std::thread myThread(serverThreadFunc);
    myThread.join();

    //  Pot_stopReading();
    Mode_end();
    //	AudioMixer_cleanup();
    return 0;
}
